/*
    Magma code related to the paper :
        "Computing images of Galois representations attached to elliptic curves", Forum of Mathematics, Sigma 4 (2016) e4 (79 pages), http://dx.doi.org/10.1017/fms.2015.33
    Copyright (c) 2015 by Andrew V. Sutherland
    You are welcome to use and/or modify this code under GPL version 2 or any later version (but please be sure to cite the paper if you use this code in your research!)
*/

load "subgroups.m";
load "distinguish.m";

det:=function(G) return {Determinant(g):g in G}; end function;
chi:=function(g) return KroneckerSymbol(Integers()!(Trace(g)^2-4*Determinant(g)),Characteristic(BaseRing(g))); end function;
rk:=function(g) M:=MatrixRing(BaseRing(g),2); return 2-Rank(M!g-M![1,0,0,1]); end function;
sig:=function(g) return <Determinant(g),Trace(g),rk(g)>; end function;
Sig:=function(G) return {sig(g):g in G}; end function;
tzr:=function(G) return #{g:g in G|Trace(g) eq 0}/#G; end function;

tchi:=function(t) return KroneckerSymbol(Integers()!(t[2]^2-4*t[1]),Characteristic(Parent(t[1]))); end function;
tord:=function(t) o:=1; a0:=2; a1:=t[2]; d:=t[1]; while d ne 1 or a1 ne 2 do a2:=t[2]*a1-a0*t[1]; a0:=a1; a1:=a2; d:=d*t[1]; o+:=1; end while; return o; end function;
tpord:=function(t) o:=1; a0:=2; a1:=t[2]; d:=t[1]; while 4*d ne a1^2 do a2:=t[2]*a1-a0*t[1]; a0:=a1; a1:=a2; d:=d*t[1]; o+:=1; end while; return o; end function;
tmat:=function(t)
    p:=Characteristic(Parent(t[1]));
    if tchi(t) eq 0 then return GL(2,p)![t[2]/2,0,0,t[2]/2]; end if;
    if tchi(t) eq 1 then
        R<x>:=PolynomialRing(GF(p));
        r:=[a[1]:a in Roots(x^2-t[2]*x+t[1])];
        return GL(2,p)![r[1],0,0,r[2]];
    end if;
    x:=t[2]/2;
    e:=GF(p)!PrimitiveRoot(p);
    y:=Sqrt((x^2-t[1])/e);
    return GL(2,p)![x,e*y,y,x];
end function;
                

IsLocallyConjugate := function(G,H1,H2)
    p:=Characteristic(BaseRing(G));
    if #H1 ne #H2 then return false; end if;
    if IsConjugate(G,H1,H2) then return true; end if;
    if #H1 mod p ne 0 or #H2 mod p ne 0 then return false; end if;
    if Min({chi(g):g in H1}) lt 0 or Min({chi(g):g in H2}) lt 0 then return false; end if;
    if Sig(H1) ne Sig(H2) or tzr(H1) ne tzr(H2) then return false; end if;
    b,pi:=IsIsomorphic(H1,H2);
    if not b then return false; end if;
    r:=PrimitiveRoot(p);
    B:=sub<GL(2,p)|[1,1,0,1],[1,0,0,r],[r,0,0,1]>;
    S1:=[x:x in Conjugates(G,H1)|x subset B];
    if #S1 eq 0 then return false; end if;
    G1 := S1[1];
    S2:=[x:x in Conjugates(G,H2)|x subset B];
    if #S2 eq 0 then return false; end if;
    G2 := S2[1];
    G1:={[g[1][1],g[2][2]]:g in G1|g[1][2] eq 0};
    G2:={[g[1][1],g[2][2]]:g in G2|g[1][2] eq 0};
    G3:={[g[2],g[1]]:g in G2};
    if G1 eq G3 then return true; end if;
    return false;
end function;
    
// Algorithm 1: compute generators for subgroup with a given signature and trace zero ratio
gfs :=function(s,zr)
    p :=Characteristic(Parent(Random(s)[1]));
    F:=GF(p);
    R<x>:=PolynomialRing(F);
    z:=F!PrimitiveRoot(p);
    GG := GL(2,p);
    // ell = 2
    if p eq 2 then
        if #s eq 1 then return sub<GG|Identity(GG)>; end if;
        if #s eq 2 and <1,0,1> in s then return sub<GG|[0,1,1,0]>; end if;
        if #s eq 2 then return sub<GG|[1,1,1,0]>; end if;
        return GG;
    end if;
    // order ell cases
    if <1,2,1> in s then
        a := [n:n in GF(p)|not IsSquare(-n)][1];
        d := z^ExactQuotient(p-1,#{t[1]:t in s});
        if #{t:t in s|tchi(t) eq -1} gt 0 then
            return sub<GG|[1,1,0,1],[0,-a,1/a,0],[1,0,0,d]>;
        else
            o:=Max({Order(Roots(x^2-t[2]*x+1)[1][1]):t in s | t[1] eq 1});
            r:= z^ExactQuotient(p-1,o);
            o:=Max({Order(t[1]):t in s});
            d:=z^ExactQuotient(p-1,o);
            t:=[t:t in s|t[1] eq d][1];
            a:=[b[1]: b in Roots(x^2-t[2]*x+t[1])];
            if #a eq 1 then a:=Append(a,a[1]); end if;
            return sub<GG|[1,1,0,1],[r,0,0,1/r],[a[1],0,0,a[2]]>;
        end if;
    end if;
    // Exceptional cases
    S:= {tpord(t):t in s};
    dets := {t[1]:t in s};
    Z:=sub<GG|[tmat(t):t in s|tchi(t) eq 0]>;
    // A4
    if S eq {1,2,3} and zr lt 3/8 then
        g:=tmat([t:t in s|tpord(t) eq 3][1]);
        h:=tmat([t:t in s|tpord(t) eq 2][1]);
        hs:=[x:x in Conjugates(GG,h)|sig(x*g) in s and sig(x^-1*g) in s];
        for h in hs do H:=sub<GG|g,h,Z>; if Sig(H) eq s and tzr(H) eq 1/4 then return H; end if; end for;
        assert false;
    end if;
    // S4
    if S eq {1,2,3,4} then
        g:=tmat([t:t in s|tpord(t) eq 3][1]);
        h:=tmat([t:t in s|tpord(t) eq 4][1]);
        hs:=[x:x in Conjugates(GG,h)|sig(x*g) in s and sig(x^-1*g) in s];
        for h in hs do H:=sub<GG|g,h,Z>; if Sig(H) eq s then return H; end if; end for;
        assert false;
    end if;
    // A5
    if S eq {1,2,3,5} then
        g:=tmat([t:t in s|tpord(t) eq 3][1]);
        h:=tmat([t:t in s|tpord(t) eq 5][1]);
        hs:=[x:x in Conjugates(GG,h)|sig(x*g) in s and sig(x^-1*g) in s];
        for h in hs do H:=sub<GG|g,h,Z>; if Sig(H) eq s then return H; end if; end for;
        assert false;
    end if;
    // trivial
    if #{tchi(t):t in s} eq 1 then return Z; end if;
    m:=Max({tpord(t):t in s});
    g:=[tmat(t):t in s|tpord(t) eq m][1];
    m:=Max({tord(t):t in s|tchi(t) eq 0});
    z:=[tmat(t):t in s|tord(t) eq m and tchi(t) eq 0][1];
    H:=sub<GG|g,z>;
    // cyclic
    if Sig(H) eq s and Abs(tzr(H)-zr) lt 1/8 then return H; end if;
    // dihedral
    e := ExactQuotient(p-1,#{h:h in H|chi(h) eq 0});
    if chi(g) eq 1 then
        gamma := GG![0,1,1,0];
        r:=GG![1,0,0,PrimitiveRoot(p)];
    else
        gamma := GG![1,0,0,-1];
        a:=GF(p^2).1;
        r:=tmat([Norm(a),Trace(a),0]);
    end if;
    G:=sub<GG|H,gamma>;
    if Sig(G) eq s then return G; else return sub<GG|H,gamma*r^e>; end if;      
    assert false;
end function;
    
verify_gfs := function(p)
    GG := GL(2,p);
    time S:=[G`subgroup:G in Subgroups(GG)];
    t0:=Cputime();
    cnt:=0;
    for i in [1..#S] do
        t1:=Cputime();
        G:=S[i];
        H:=gfs(Sig(G),tzr(G));
        //assert Sig(H) eq Sig(G);
        assert IsLocallyConjugate(GG,G,H);
        cnt+:=1;
        t2:=Cputime();
        if t2-t1 gt 1 then print cnt, "of", #S, "for p =", p, "in", t2-t1,"secs"; end if;
    end for;
    print "Verified", cnt, "of", #S, "for p =", p, "in", t2-t0,"sec";
    return true;
end function;

// Step 2 of Algorithm 4
TorsionRank:=function(E,N,ell)
    e := Valuation(N,ell);
    if e lt 2 then return e; end if;
    if #BaseRing(E) mod ell ne 1 then return 1; end if;
    M:=ExactQuotient(N,ell^e);
    O := E!0;
    // Get a point P of order ell
    P:=O;
    while P eq O do P:= M*Random(E); end while;
    a:=1; X:=ell*P; while X ne O do a+:=1; P:= X; X*:=ell; end while;
    if a eq e then return 1; end if;
    // generate random points Q of ell-power order until we either find a basis for the ell-torsion or prove the ell-power torsion is cyclic.
    while true do
        Q:=M*Random(E);
        if Q eq E!0 then continue; end if;
        a:=1; X:=ell*Q; while X ne O do a+:=1; Q:= X; X*:=ell; end while;
        if a eq e then return 1; end if;
        if WeilPairing(P,Q,ell) ne 1 then return 2; end if;
    end while;
end function;

// Algorithm 4
//   E is an elliptic curve over a finite field
//   S is a prime ell or list of primes ell
//   t is optional, it allows the caller to pass in the trace if known
EpSigs:=function(E,S:t:=TraceOfFrobenius(E))
    if Type(S) ne SeqEnum then S:=[S]; end if;
    q:=#BaseRing(E); 
    T:=[*<>:ell in S*];
    for i:=1 to #S do
        ell:=S[i];
        if IsDivisibleBy(q,ell) then continue; end if;
        r:=TorsionRank(E,q+1-t,ell);
        T[i]:=<GF(ell)!q, GF(ell)!t, r>;
    end for;
    return T;
end function;

// Variant of Algorithm 5
//   E is an elliptic curve over a number field
//   S is a prime ell or list of primes el
//   B is a bound on the norms of the primes p to check (it will automatically be increased if it appears to be too low)
//   K is an optional field of definition, if different from the field over whih E is defined
//      if K is specified E will be base-changed to K, but the algorithm can take advantage of the fact that E is
//      defined over a proper subfield (currently only relevant when E is defined over Q)
GaloisImages:=function(E,S,B:K:=BaseRing(E))
    if Type(S) ne SeqEnum then S:=[S]; end if;
    for ell in S do assert IsPrime(ell); end for;
    D:=Integers()!Norm(Discriminant(E));
    fK:=DefiningPolynomial(K);
    d:=Integers()!Discriminant(fK);
    M:=[0:ell in S];
    T:=[*{}:ell in S*];
    cnt:=[0:ell in S];
    zcnt:=[0:ell in S];
    if B lt 1000 then B:=1000; end if;
    P:=PrimesInInterval(1,B);
    K0:=BaseRing(E);
    if K0 eq Rationals() then
        // For curves defined over Q this is much faster than computing traces individually
        Qtraces:=TracesOfFrobenius(E,B);
    else
        // we only optimize base changes for E/Q
        E:=ChangeRing(E,K);
        K0:=K;
        OK:=RingOfIntegers(K);
    end if;
    for i:=1 to #P do
        p:=P[i];
        if IsDivisibleBy(D,p) or IsDivisibleBy(d,p) then continue; end if;
        if K0 eq Rationals() then
            // For base changes we only look at degree-1 primes
            n:=#Roots(ChangeRing(fK,GF(p)));
            if n eq 0 then continue; end if;
            t:=Qtraces[i];
            Ep:=ChangeRing(E,GF(p));
            s:=EpSigs(Ep,S:t:=t);
            for j:=1 to #S do
                if S[j] eq p then continue; end if;
                if not s[j] in T[j] then
                    Include(~T[j], s[j]);
                    M[j]:=i;
                end if;
                cnt[j]+:=n;
                if s[j][2] eq 0 then zcnt[j]+:=n; end if;
            end for;
        else
            A:=Factorization(ideal<OK|p>);
            for a in A do
                pp:=a[1];
                if Norm(pp) le B then
                    Ep:=Reduction(E,pp);
                    s:=EpSigs(Ep,S);
                    for j:=1 to #S do
                        if S[j] eq p then continue; end if;
                        if not s[j] in T[j] then
                            Include(~T[j], s[j]);
                            M[j]:=i;
                        end if;
                        cnt[j]+:=1;
                        if s[j][2] eq 0 then zcnt[j]+:=1; end if;
                    end for;
                end if;
            end for;
        end if;
    end for;
    if Min(cnt) le Max(S)^2 or Max(M) gt #P-500 then
        printf "B=%o appears to be to small, doubling it...\n", B;
        return $$(E,S,2*B:K:=K);
    end if;
    G:=[**];
    for i:=1 to #S do
        z :=zcnt[i]/cnt[i];
        try
            G[i]:=GL2SubgroupLabel(gfs(T[i],z));
        catch e
            printf "B=%o is too small, doubling it...\n", B;
            return $$(E,S,2*B);
        end try;
        C:=GL2LocalConjugates(G[i]);
        if #C ne 1 then
            assert #C eq 2;
            C:=[c:c in C];
            TD:=[GroupTorsionDegree(GL2SubgroupFromLabel(c)):c in C];
            if TD[1] eq TD[2] then
                td:=CurveTorsionDegree(E,S[i]);
                assert td in TD;
                if td eq TD[1] then G[i]:=C[1]; else G[i] := C[2]; end if;
            else
                G[i]:=Sprintf("%o or %o", C[1], C[2]);
            end if;
        end if;
    end for;
    return G;
end function;









    
